from enum import Enum

import numpy as np


class LogLevel(Enum):
    DEBUG = 0,
    TRACE = 1,
    WARNING = 2,
    ERROR = 3,
    FATAL = 4


class Logger:

    def __init__(self, log_level: LogLevel):
        self.log_level = log_level

    def log(self, log_level: LogLevel, message: str):
        if log_level >= self.log_level:
            print(message)


class CollisionSet:

    def __init__(self, mini_batch_size):
        self.data = np.zeros(mini_batch_size)
        self._mini_batch_size = mini_batch_size

    def set_collision(self, index):
        self.data[index] = 1

    def crash_list(self):
        return self.data == 1

    def collision_at(self, index):
        return self.data[index] == 1

    def full_collision(self):
        return self.data == np.ones(len(self.data))

    def clear(self, mini_batch_size):
        """
        mini_batch_size can be smaller than self.mini_batch_size
        :param mini_batch_size:
        :return:
        """
        self.data = np.zeros(mini_batch_size)



import os
from abc import ABC, abstractmethod
from time import time
from gui.components import Label, TimedLabel, TimedLabelQueue, ArrowKeysDisplay
import numpy as np
import pygame

class Simulation(ABC):

    def __init__(self,
                 debug=True,
                 fps=None,
                 num_episodes=None,
                 caption: str = None,
                 car=None,
                 track_offset=(0, 0),
                 screen_size=(1400, 800),
                 track_size=(1400, 800)):
        """

        Simulation should hold only the information relevant to the actual simulation and not any information
        about the agent driving (the ai or human)

        Rewards should be initialized within the agent itself and not the simulator

        :param debug: whether or not sensors/etc. will be shown
        :param fps: None if simulation is not run based on fps (speed of while-loop) otherwise fps of simulator
        :param num_episodes: the number of episodes (crashes) before the simulation dies
                            - None if the simulator runs forever
        """
        pygame.init()

        # private attributes
        self._screen_dim = screen_size
        self._track_dim = track_size
        self._fps = fps
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # track information
        self.start_pos = None
        self._track_border = None
        self.track_border_width = None
        self.track_border_height = None
        self.border_mask = None
        self._track_bg = None
        self._track_rewards = None
        self.rewards_mask = None
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        self._clock = pygame.time.Clock()
        self._iteration_num = 1
        self._debug = debug
        self._max_episodes = num_episodes
        self._caption = caption
        self._track_offset = track_offset

        # a list of rects that have been updated since the last screen refresh
        self._rect_changed = []
        # helper attribute for calculating the actual fps
        self.current_timestamp = None
        self._calc_fps = 0

        # public attributes
        self.window = pygame.display.set_mode(self._screen_dim)
        # labels
        self.fps_label = None
        self.iteration_count_label = None
        self.car_label = None
        self.car_speed_label = None
        self.odometer_label = None
        self.label_manager = TimedLabelQueue(self.window)
        self.arrow_display = ArrowKeysDisplay(
            unselected_color=(255, 255, 255),
            selected_color=(255, 0, 255),
            border=(0, 0, 0),
            scale=1,
            position=None
        )
        # this value should be overridden by child class
        self.start_pos = (0, 0)
        # initialize car later
        self.car = car
        self._car_top_distance = 0

        # handle init
        self.init_display()
        self.init_iteration_count_label()
        self.init_fps_label()
        self.init_car_start_pos()
        self.init_car_label()
        self.init_car_stat_labels()
        self.convert_images()  # initializes the track

    def init_car_stat_labels(self):
        self.odometer_label = Label(position=(600, self._screen_dim[1] - 50), text=" Pixels", size=40, font=None,
                                    color=(255, 255, 255), refresh_count=None, background=None, anti_alias=True)
        # Position is arbitrary as it will be moved
        self.car_speed_label = Label(position=(0, 0), text=" PPF", size=40, font=None,
                                 color=(0, 0, 0), refresh_count=None, background=None, anti_alias=True)

    @abstractmethod
    def init_car_start_pos(self):
        """
        sets the start position of the car
        :return:
        """
        pass

    @abstractmethod
    def init_track(self) -> (str, str, str):
        """
        Should set the images of the track (paths to the images):
        called in the constructor of the simulation class
        - track border
        - track bg
        - track rewards
        :return: the path to the tracks in the order 'border, background (design), rewards'
        """
        pass

    def convert_images(self):
        """
        called in constructor for converting images
        :return:
        """
        border, track, rewards = self.init_track()
        self._track_border = pygame.image.load(border)
        self._track_border = pygame.transform.smoothscale(self._track_border, self._track_dim).convert_alpha()
        self._track_bg = pygame.image.load(track).convert()
        self._track_bg = pygame.transform.smoothscale(self._track_bg, self._track_dim).convert()
        self._track_rewards = pygame.image.load(rewards)
        self._track_rewards = pygame.transform.smoothscale(self._track_rewards, self._track_dim).convert_alpha()

        if self._track_border is not None:
            self.track_border_width = self._track_border.get_width()
            self.track_border_height = self._track_border.get_height()
        if border is not None:
            self.border_mask = pygame.mask.from_surface(self._track_border)
        if rewards is not None:
            self.rewards_mask = pygame.mask.from_surface(self._track_rewards)

    def get_vehicle_offset(self):
        return None if self.car is None else (self.car.image.get_width() / 2, self.car.image.get_height() / 2)

    def get_vehicle_image_position(self):
        """
        :return: The absolute position of the image of the vehicle (in relation to the window)
        """
        return np.array((self.car.velocity.x + (self.car.image.get_width() / 2) + 12,
                  self.car.velocity.y + (self.car.image.get_height() / 2) + 12))

    def init_display(self):
        if self._caption is None:
            self._caption = "Racing Simulation"

    def init_fps_label(self):
        self.fps_label = Label((10, 10), "FPS: 0", size=30, font=None, color=(0, 0, 0), background=None,
                               anti_alias=True)

    def init_car_label(self):
        self.car_label = Label((10, self._screen_dim[1] - 40), "Longest Drive: ", size=30, font=None, color=(0, 0, 0),
                               background=(255, 255, 255), anti_alias=True)

    def init_iteration_count_label(self):
        self.iteration_count_label = Label((1100, 10), "Iteration: ", size=30, font=None, color=(0, 0, 0),
                                   background=(255, 255, 255), anti_alias=True)

    def simulate(self):
        """
        main 'game-loop' for simulation
        :return: None
        """
        # - - - - - - - - - - - - - - - - - - - - - - - - - -
        print("Begin simulation init...")
        run = True
        pygame.display.set_caption(self._caption)
        if self.car is not None:
            print("Initializing car image conversion...")
            self.car.image = self.car.image.convert()
            self.car.reset(simulation=self)
        print("Done!")
        # - - - - - - - - - - - - - - - - - - - - - - - - - -
        while run:
            if self._fps is not None and run:
                self._clock.tick(self._fps)
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    run = False
                    if self.car is not None:
                        print("Saving Car...")
                        self.car.save_car()
                    break
            t = self.current_timestamp
            self.current_timestamp = time()
            if t is not None:
                self.calculate_fps(self.current_timestamp - t)
            keys_pressed = pygame.key.get_pressed()
            self.update_display(keys_pressed)

    def calculate_fps(self, time_elapsed):
        # convert to seconds (from milliseconds)
        t = time_elapsed
        # save to attribute
        self._calc_fps = round(1 / t)

    def update_display(self, keys_pressed):
        """
        :param keys_pressed:
        :return:
        """
        self.window.fill((0, 0, 0))
        if self._track_bg is not None:
            self.window.blit(self._track_bg, self._track_offset)
        if self._track_rewards is not None and self._debug:
            self.window.blit(self._track_rewards, self._track_offset)
        self.car.update_sensors(self.window, self)
        reward = self.handle_reward()
        if reward:
            pass
        collision = self.handle_collision()
        if collision:
            self.reset()
        self.car.step(reward, collision, keys_pressed)
        self.car.update(simulation=self)
        self.update_debug_display(reward, collision)
        self.update_and_display_labels()
        self.update_and_display_arrow_display()
        self.op_display()

    @abstractmethod
    def update_and_display_arrow_display(self):
        pass

    def update_and_display_labels(self):
        self.fps_label.append_text(str(self._calc_fps), self._calc_fps / 2)
        self._car_top_distance = max(self._car_top_distance, self.car.odometer)
        self.car_label.append_text(str(round(self._car_top_distance)), refresh_count=None)
        self.car_speed_label.append_text(str(round(self.car.velocity.speed)), self._calc_fps / 2, append_to_front=True)
        self.odometer_label.append_text(str(round(self.car.odometer)), refresh_count=None, append_to_front=True)
        self.iteration_count_label.append_text(str(self._iteration_num))
        self.iteration_count_label.render(self.window)
        self.car_label.render(self.window)
        self.car_speed_label.render(self.window, position=(self.car.velocity.x + 10, self.car.velocity.y - 40))
        self.odometer_label.render(self.window)
        self.fps_label.render(self.window)
        self.label_manager.render()

    # TODO implement
    def op_display(self):
        """
        optimize what elements are getting displayed to the screen
        if no elements are updated the screen shouldn't update
        :return:
        """
        # fixme -> add check to see what elements have been updated
        # pygame.display.update(self._rect_changed)
        pygame.display.update()

    def update_debug_display(self, reward: bool, collision: bool):
        """
        if debug is active then display elements
        :return:
        """
        pass

    def reset(self):
        """
        Called whenever the car crashes and the simulation starts over
        :return:
        """
        self._iteration_num += 1
        self.car.reset(simulation=self)

    # - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Handlers for display/collision detection
    # - - - - - - - - - - - - - - - - - - - - - - - - - -

    def handle_collision(self) -> bool:
        """
        Only works if the track_border is not None
        :return: whether the vehicle hit a wall
        """
        if self._track_border is not None \
                and self.car is not None \
                and self.car.current_image is not None:
            car_mask = pygame.mask.from_surface(self.car.image)
            x, y = (self.car.velocity.x + (.5 * self.car.image.get_width()),
                    self.car.velocity.y + (.5 * self.car.image.get_height()))
            col = self.border_mask.overlap(car_mask, (x, y))
            if col is not None:
                return True
        return False

    def handle_reward(self) -> bool:
        """
        Only works if the track_rewards is not None
        :return: whether the vehicle is touching a reward
        """
        if self._track_rewards is not None \
                and self.car is not None \
                and self.car.current_image is not None:
            car_mask = pygame.mask.from_surface(self.car.image)
            car_pos = self.get_vehicle_image_position()
            col = self.rewards_mask.overlap(car_mask, (car_pos[0], car_pos[1]))
            if col is not None:
                return True
        return False

    # - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Display for application aesthetics and debugging
    # - - - - - - - - - - - - - - - - - - - - - - - - - -

    def display_sensor_values(self, start_pos: (int, int)):
        pass



from abc import ABC
from time import time

import numpy as np
import pygame
from pygame import K_DOWN

from gui.components import TimedLabelQueue, TimedLabel, Label
from agent import Agent
from example import Car
import os
from vehicle import SensorBuilder, Sensor
from utils import CollisionSet


class GeneticCar(Car):

    def __init__(self, driver, car_number, debug=False, acceleration_multiplier=.5, normalize=True):
        super(GeneticCar, self).__init__(driver, debug, acceleration_multiplier, normalize)
        self.number = 0
        self.car_number = car_number
        self.parents = []
        self.collision = False

    @staticmethod
    def generate_cars(batch_size, params=None):
        if params is None:
            params = {
                'acceleration_multiplier': .5,
                'normalize': True,
                'debug': True
            }
        return np.array([GeneticCar(
            driver=None, car_number=num, **params
        ) for num in range(batch_size)])

    def reset(self, simulation):
        """
        - Resets the car properties, so it is ready for another episode
        TODO - same implementation of superclass
        :return: None
        """
        self.velocity.reset_velocity(
            x=simulation.start_pos[0],
            y=simulation.start_pos[1],
            angle=180,
            speed=0
        )
        self.collision = False
        self.odometer = 0

    def __add__(self, other):
        """
        Perform cross over with other
        :param other:
        :return:
        """
        return self.driver + other.driver

    def __lt__(self, other):
        return self.odometer < other.odometer

    def __gt__(self, other):
        return self.odometer > other.odometer

    def __eq__(self, other):
        return self.odometer == other.odometer

    def update_sensors(self, window, simulation):
        """
        Calculate the values for the sensors of the vehicle
        :param window: the pygame window
        :param simulation: the simulation that the car is functioning in
        :return: None
        """
        # get (x1,y1,x2,y2) tuples for all sensor positions
        s = [sensor.update(car=self, window=window, simulation=simulation) for sensor in self.sensors]
        if self._debug:
            self.display_sensor(car_pos=GeneticAlgorithmSimulation.get_vehicle_image_position(self), window=window)

    def update(self, simulation):
        """
        update necessary data and send back to simulation
        :param simulation: the simulation the vehicle exists in
        :return: None
        """
        window = simulation.window
        # account for reoccurring events (such as velocity update)
        if not self.collision:
            self.update_pos()
            self.odometer += self.velocity.speed  # update odometer as the distance moved each step
        self.blit_rotate_center(window, (self.velocity.x, self.velocity.y))
        if self.collision:
            # TODO - Magic number is simply an offset, replace with calculated value
            pygame.draw.line(window, (255, 0, 0), (self.velocity.x + 15, self.velocity.y),
                             (self.velocity.x + self.current_image.get_width() + 15,
                              self.velocity.y + self.current_image.get_width()), width=10)
            pygame.draw.line(window, (255, 0, 0), (self.velocity.x + 15, self.velocity.y + self.current_image.get_width()),
                             (self.velocity.x + self.current_image.get_width() + 15, self.velocity.y), width=10)


class GeneticCarSet:

    def __init__(self, cars, mini_batch_size):
        """
        Single entity to perform operations concerning all cars
        :param cars: a numpy array of cars
        :param mini_batch_size: the number of cars in the mini-batch
        """
        self.cars = cars
        self.sensor_builder = None
        self.batch_size = len(cars)
        self.car_offset = self.cars[0].image.get_width() / 2, self.cars[0].image.get_height() / 2
        self.mini_batch_size = mini_batch_size
        # batch state
        self.batch_results_index = 0
        self.batch_results = np.zeros(self.batch_size)
        # mini-batch state
        self.mini_batch_index = self.mini_batch_size
        self.mini_batch = self.cars[:self.mini_batch_size]
        self.collision_set = CollisionSet(mini_batch_size=mini_batch_size)
        self.speed_zero_counter = np.zeros(mini_batch_size)
        self._speed_zero_max = 50
        # other
        # simulation must be set prior to running simulation.simulate()
        self.simulation = None

    def initialize(self):
        if self.cars is not None:
            print("Initializing car image conversion...")
            for car in self.cars:
                car.image = car.image.convert()

    def weed_the_weak(self):
        for i, car in enumerate(self.mini_batch):
            if car.velocity.speed <= 1:
                if self.speed_zero_counter[i] >= self._speed_zero_max:
                    self.collision_set.set_collision(i)
                    car.collision = True
                else:
                    self.speed_zero_counter[i] += 1

    def get_external_inputs(self):
        return self.cars[0].get_external_inputs()

    def initialize_sensors(self, sensor_batch):
        for car, sensors in zip(self.cars, sensor_batch):
            car.sensors = sensors

    def initialize_drivers(self, drivers, simulation):
        for car, driver in zip(self.cars, drivers):
            car.driver = driver
            car.reset(simulation=simulation)

    def reset_all(self):
        for car in self.cars:
            car.reset(self.simulation)

    def handle_reset(self, simulation):
        """
        reset the mini_batch if all cars are collided
        :return: None
        """
        if self.collision_set.full_collision().all():
            self.reset_mini_batch(simulation)

    def reset_mini_batch(self, simulation):
        """
        reset the collision_set and queue the next set of cars
        :return:
        """
        print("Batch Index:", self.mini_batch_index)
        print("Longest Distance:", simulation.longest_distance)
        self.speed_zero_counter = np.zeros(self.mini_batch_size)
        for car in self.mini_batch:
            self.batch_results[self.batch_results_index] = car.odometer
            self.batch_results_index += 1
            if car.odometer > simulation.longest_distance:
                simulation.best_driver = car.driver
                simulation.longest_distance = car.odometer
            car.reset(simulation)

        # If the batch is done
        if self.mini_batch_index >= len(self.cars):
            self.clean_up_iteration(simulation)
            self.batch_results_index = 0
            self.mini_batch_index = 0
        next_index = self.mini_batch_index + self.mini_batch_size
        # grab the next mini_batch of cars
        self.mini_batch = self.cars[
                          self.mini_batch_index: next_index if next_index < len(self.cars) else len(self.cars)]
        self.mini_batch_index += self.mini_batch_size
        self.collision_set.clear(len(self.mini_batch))

    def clean_up_iteration(self, simulation):
        """
        called when the entire batch is complete
        :return:
        """
        simulation.clean_up_iteration()
        s = np.argsort(self.batch_results)
        parent_cars = self.cars[s[-2:]]
        driver1, driver2 = parent_cars[0].driver, parent_cars[1].driver
        # Note: new_drivers includes driver1 and driver2
        new_drivers = driver1.cross_over_mutation(driver2, total_batch_size=self.batch_size)
        self.initialize_drivers(new_drivers, simulation)

    def update_sensors(self, window, simulation):
        for i, car in enumerate(self.mini_batch):
            if not self.collision_set.collision_at(i):
                car.update_sensors(window, simulation)

    def step_mini_batch(self, keys_pressed):
        force_reset = False
        if keys_pressed[K_DOWN]:
            force_reset = True
        for i, car in enumerate(self.mini_batch):
            if force_reset:
                car.collision = True
                self.collision_set.set_collision(i)
            if not self.collision_set.collision_at(i):
                car.step(
                    reward=False,
                    collision=False,
                    keys_pressed=keys_pressed
                )

    def update_mini_batch(self, simulation):
        for i, car in enumerate(self.mini_batch):
            car.update(simulation)


class GeneticAlgorithmSimulation:

    def __init__(self,
                 debug=True,
                 fps=None,
                 num_episodes=None,
                 cars: GeneticCarSet = None,
                 track_offset=(0, 0),
                 screen_size=(1400, 800),
                 track_size=(1400, 800),
                 batch_size=100,
                 mini_batch_size=3,
                 caption="Genetic Algorithm"
                 ):

        """

                Simulation should hold only the information relevant to the actual simulation and not any information
                about the agent driving (the ai or human)

                Rewards should be initialized within the agent itself and not the simulator

                :param debug: whether or not sensors/etc. will be shown
                :param fps: None if simulation is not run based on fps (speed of while-loop) otherwise fps of simulator
                :param num_episodes: the number of episodes (crashes) before the simulation dies
                                    - None if the simulator runs forever
                """
        pygame.init()

        # private attributes
        self._screen_dim = screen_size
        self._track_dim = track_size
        self._fps = fps
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # track information
        self.start_pos = None
        self._track_border = None
        self.track_border_width = None
        self.track_border_height = None
        self.border_mask = None
        self._track_bg = None
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        self._clock = pygame.time.Clock()
        self.iteration_num = 1
        self.longest_distance = 0
        self._debug = debug
        self._max_episodes = num_episodes
        self._caption = caption
        self._track_offset = track_offset

        # a list of rects that have been updated since the last screen refresh
        self._rect_changed = []
        # helper attribute for calculating the actual fps
        self.current_timestamp = None
        self._calc_fps = 0

        # public attributes
        self.window = pygame.display.set_mode(self._screen_dim)
        # labels
        self.fps_label = None
        self.iteration_count_label = None
        self.car_label = None
        self.label_manager = TimedLabelQueue(self.window)
        # this value should be overridden by child class
        self.start_pos = (0, 0)
        self.best_driver = None

        # handle init
        self.init_display()
        self.init_iteration_count_label()
        self.init_fps_label()
        self.init_car_start_pos()
        self.init_car_label()
        self.convert_images()  # initializes the track
        self.mini_batch_size = mini_batch_size
        self.current_index = mini_batch_size
        # stores the odometer readings from cars that have "competed"
        self.batch_size = batch_size
        self.cars: GeneticCarSet = cars
        self.car_death_count = 0

    def convert_images(self):
        """
        called in constructor for converting images
        :return:
        """
        border, track = self.init_track()
        self._track_border = pygame.image.load(border)
        self._track_border = pygame.transform.smoothscale(self._track_border, self._track_dim).convert_alpha()
        self._track_bg = pygame.image.load(track).convert()
        self._track_bg = pygame.transform.smoothscale(self._track_bg, self._track_dim).convert()

        if self._track_border is not None:
            self.track_border_width = self._track_border.get_width()
            self.track_border_height = self._track_border.get_height()
        if border is not None:
            self.border_mask = pygame.mask.from_surface(self._track_border)

    def get_vehicle_offset(self):
        return None if self.cars is None else self.cars.car_offset

    @staticmethod
    def get_vehicle_image_position(car):
        """
        :return: The absolute position of the image of the vehicle (in relation to the window)
        """
        return np.array((car.velocity.x + (car.image.get_width() / 2) + 12,
                         car.velocity.y + (car.image.get_height() / 2) + 12))

    def init_display(self):
        if self._caption is None:
            self._caption = "Racing Simulation"

    def init_fps_label(self):
        self.fps_label = Label((10, 10), "FPS: 0", size=30, font=None, color=(0, 0, 0), background=None,
                               anti_alias=False)

    def init_car_label(self):
        self.car_label = Label((10, self._screen_dim[1] - 40), "Speed: ", size=30, font=None, color=(0, 0, 0),
                               background=(255, 255, 255), anti_alias=False)

    def init_iteration_count_label(self):
        self.iteration_count_label = Label((1100, 10), "Iteration: ", size=30, font=None, color=(0, 0, 0),
                                           background=(255, 255, 255), anti_alias=False)

    def simulate(self):
        """
        main 'game-loop' for simulation
        :return: None
        """
        # - - - - - - - - - - - - - - - - - - - - - - - - - -
        print("Begin simulation init...")
        run = True
        pygame.display.set_caption(self._caption)
        self.cars.initialize()
        print("Done!")
        # - - - - - - - - - - - - - - - - - - - - - - - - - -
        while run:
            if self._fps is not None and run:
                self._clock.tick(self._fps)
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    run = False
                    if self.best_driver is not None:
                        path = os.path.join("assets", "models")
                        if not os.path.exists(path):
                            os.mkdir(path)
                        self.best_driver.save_model(os.path.join("assets", "models"))
                    break
            t = self.current_timestamp
            self.current_timestamp = time()
            if t is not None:
                self.calculate_fps(self.current_timestamp - t)
            keys_pressed = pygame.key.get_pressed()
            self.update_display(keys_pressed)

    def calculate_fps(self, time_elapsed):
        # convert to seconds (from milliseconds)
        t = time_elapsed
        # save to attribute
        self._calc_fps = round(1 / t)

    def update_display(self, keys_pressed):
        """
        :param keys_pressed:
        :return:
        """
        self.window.fill((0, 0, 0))
        if self._track_bg is not None:
            self.window.blit(self._track_bg, self._track_offset)
        self.cars.update_sensors(self.window, self)
        # update the collision set to determine if all cars in the minibatch have 'crashed'
        self.handle_collision()
        # if a car "stalls" call it a collision
        self.cars.weed_the_weak()
        # Only call reset when all cars have 'died'
        self.cars.handle_reset(self)
        self.cars.step_mini_batch(keys_pressed=keys_pressed)
        self.cars.update_mini_batch(simulation=self)
        self.update_and_display_labels()
        self.op_display()

    def handle_collision(self) -> None:
        """
        Only works if the track_border is not None
        :return: whether the vehicle hit a wall
        """
        for i, car in enumerate(self.cars.mini_batch):
            if self._track_border is not None \
                    and car is not None \
                    and car.current_image is not None:
                car_mask = pygame.mask.from_surface(car.image)
                # TODO remove manual offset and use function call
                x, y = (car.velocity.x + (.5 * car.image.get_width()),
                        car.velocity.y + (.5 * car.image.get_height()))
                col = self.border_mask.overlap(car_mask, (x, y))
                if col is not None:
                    self.cars.collision_set.set_collision(i)
                    car.collision = True
            else:
                raise Exception("Track has not been properly initialized.")

    def update_and_display_labels(self):
        self.iteration_count_label.append_text(str(self.iteration_num))
        self.iteration_count_label.render(self.window)
        self.label_manager.render()

    # TODO implement
    def op_display(self):
        """
        optimize what elements are getting displayed to the screen
        if no elements are updated the screen shouldn't update
        :return:
        """
        # fixme -> add check to see what elements have been updated
        # pygame.display.update(self._rect_changed)
        pygame.display.update()

    def update_debug_display(self, reward: bool, collision: bool):
        """
        if debug is active then display elements
        :return:
        """
        pass

    def reset(self):
        """
        resets the minibatch to begin the next one
        Called whenever all cars in mini_batch crash and the next mini_batch begins
        :return:
        """
        # restart the collision set
        self.cars.reset_mini_batch()

    def clean_up_iteration(self):
        """
        - Clean up the current iteration for the next one
        - Find the highest ranking cars, then crossover, then mutate
        :return:
        """
        self.iteration_num += 1

    def init_car_start_pos(self):
        """
        sets the start position of the car
        :return:
        """
        self.start_pos = (875, 100)

    def init_track(self) -> (str, str):
        """
        Should set the images of the track (paths to the images):
        called in the constructor of the simulation class
        - track border
        - track bg
        - track rewards
        :return: the path to the tracks in the order 'border, background (design), rewards'
        """
        return \
            os.path.join("assets", "track-border.png"), \
            os.path.join("assets", "track.png")


class GeneticAlgorithmDriver(Agent, ABC):

    def __init__(self, num_inputs, num_outputs, driver_id, epsilon):
        super().__init__(num_inputs, num_outputs)
        # weight initialization range of [-1.5, 1.5)
        self.w1 = 3 * np.random.random((num_inputs, 32)) - 1.5
        self.w2 = 3 * np.random.random((32, 16)) - 1.5
        self.w3 = 3 * np.random.random((16, num_outputs)) - 1.5
        self.driver_id = str(driver_id)
        self.epsilon = epsilon

    def forward(self, input_arr: np.array):
        # input shape is (1, num_input)
        pass1 = np.matmul(input_arr, self.w1)  # pass1 has shape of (1, 32)
        pass2 = np.matmul(pass1, self.w2)  # pass2 has shape of (1, 16)
        pass3 = np.matmul(pass2, self.w3)  # pass3 has shape of (1, num_outputs)
        return np.tanh(pass3)  # shape is still (1, num_outputs)

    def update(self, inputs, reward_collision=False, wall_collision=False, keys_pressed=None) -> list[int]:
        """
        - Given input from the simulation make a decision
        :param wall_collision: whether the car collided with the wall
        :param reward_collision: whether the car collided with a reward
        :param inputs: sensor input as a numpy array
        :param keys_pressed: a map of pressed keys
        :return direction: int [0 - num_outputs)
        """
        return [np.argmax(self.forward(inputs))]

    def cross_over_mutation(self, other, total_batch_size):
        """
        1. cross over self and other
        2. generate {total_batch_size - 2 (self and other)} number of mutations
        3. return the list of drivers as a numpy array, with the first two being the parents
        :param other:
        :return:
        """
        child_driver = self.cross_over(other)
        mutations = child_driver.mutate(total_batch_size - 2)
        return np.array([self, other] + mutations)

    def cross_over(self, other):
        """
        Single point crossover
        :param other:
        :return:
        """
        child_driver = GeneticAlgorithmDriver(self.num_inputs, self.num_outputs,
                                              "({} & {})".format(self.driver_id, other.driver_id), self.epsilon)
        cross_over_point = np.random.randint(self.w1.shape[1])
        child_driver.w1 = np.hstack((self.w1[:, :cross_over_point], other.w1[:, cross_over_point:]))
        cross_over_point = np.random.randint(self.w2.shape[1])
        child_driver.w2 = np.hstack((self.w2[:, :cross_over_point], other.w2[:, cross_over_point:]))
        cross_over_point = np.random.randint(self.w3.shape[1])
        child_driver.w3 = np.hstack((self.w3[:, :cross_over_point], other.w3[:, cross_over_point:]))
        return child_driver

    def mutate(self, num_mutations):
        """
        :param num_mutations: the number of mutations to create
        :return: a list of GeneticAlgorithmDriver mutations
        """
        mutations = []
        for i in range(num_mutations):
            mutation = GeneticAlgorithmDriver(self.num_inputs, self.num_outputs,
                                              driver_id="{}_{}".format(self.driver_id, i), epsilon=self.epsilon)
            mutation_arr = self.generate_mutation_arr()  # generate mutations and masks
            # read following as add mutations to self.w1 at the positions where the mask is less than epsilon
            mutation.w1 = self.w1 + mutation_arr[0][0] * (mutation_arr[0][1] < self.epsilon)
            mutation.w2 = self.w2 + mutation_arr[1][0] * (mutation_arr[1][1] < self.epsilon)
            mutation.w3 = self.w3 + mutation_arr[2][0] * (mutation_arr[2][1] < self.epsilon)
            mutations.append(mutation)
        return mutations

    def generate_mutation_arr(self):
        """
        :return: list([(mutation array, mask array), ...])
        """
        ret = []
        shapes = [self.w1.shape, self.w2.shape, self.w3.shape]
        for i in shapes:
            # range = number of layers
            ret.append((np.random.random(i) - .5, np.random.random(i)))
        return ret

    @staticmethod
    def generate_drivers(num_drivers, num_inputs, num_outputs, epsilon, load_latest=False):
        """
        create a list of random Drivers
        :param num_inputs:
        :param num_outputs:
        :param num_drivers:
        :return:
        """
        if load_latest:
            parent = GeneticAlgorithmDriver(num_inputs, num_outputs, driver_id=0, epsilon=epsilon)
            parent.load_model(os.path.join("assets", "models"))
            mutations = parent.mutate(num_drivers-1)
            return np.array([parent] + mutations)
        return np.array([GeneticAlgorithmDriver(num_inputs, num_outputs, driver_id=identifier, epsilon=epsilon)
                         for identifier in range(num_drivers)])

    def save_model(self, path):
        """
        - Save the brain of the agent to some file (or don't)
        :param path: the path to the model
        :return: None
        """
        path_name = os.path.join(path, "latest_genetic")
        if not os.path.exists(path_name):
            os.mkdir(path_name)
        np.save(os.path.join(path_name, "w1"), self.w1)
        np.save(os.path.join(path_name, "w2"), self.w2)
        np.save(os.path.join(path_name, "w3"), self.w3)

    def load_model(self, path):
        """
        - Load the brain of the agent from some file (or don't)
        :param path: the path to the model
        :return: None
        """
        path_name = os.path.join(path, "latest_genetic")
        self.w1 = np.load(os.path.join(path_name, "w1.npy"))
        self.w2 = np.load(os.path.join(path_name, "w2.npy"))
        self.w3 = np.load(os.path.join(path_name, "w3.npy"))


def main():
    # the number of cars within a particular batch
    BATCH_SIZE = 100
    # the number of cars to render on the screen at once
    MINI_BATCH_SIZE = 20
    # step 1: generate cars, and create a new car set
    genetic_cars = GeneticCar.generate_cars(BATCH_SIZE, params=None)
    car_set = GeneticCarSet(genetic_cars, MINI_BATCH_SIZE)
    # step 2: generate genetic simulation
    simulation = GeneticAlgorithmSimulation(
        debug=True,
        fps=None,
        num_episodes=None,
        cars=car_set,
        track_offset=(0, 0),
        batch_size=BATCH_SIZE,
        mini_batch_size=3,
        caption="Genetic Algorithm Simulation"
    )
    # step 3: generate sensors and attach to the cars
    sb = SensorBuilder(
        depth=500,
        sim=simulation,  # ignore warning, simulations are compatible
        default_value=None,
        color="random",
        width=2,
        pointer=True
    )
    sensor_batch = sb.generate_sensors(sensor_range=(-90, 90, 10), num_sets=BATCH_SIZE)
    # Attach the sensors to the cars
    car_set.initialize_sensors(sensor_batch=sensor_batch)
    # calculate inputs and output size
    num_inputs = car_set.get_external_inputs() + sb.num_sensors
    num_outputs = Car.get_num_outputs()
    # step 4: initialize the drivers given the input/output numbers and put them in the cars
    initial_drivers = GeneticAlgorithmDriver.generate_drivers(BATCH_SIZE, num_inputs, num_outputs, epsilon=.50, load_latest=True)
    car_set.initialize_drivers(drivers=initial_drivers, simulation=simulation)
    # step 5: simulate!
    simulation.simulate()


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod


class Agent(ABC):

    def __init__(self, num_inputs, num_outputs):
        self.num_inputs = num_inputs
        self.num_outputs = num_outputs

    """
    - Abstract agent class
    """
    @abstractmethod
    def update(self, inputs, reward_collision=False, wall_collision=False, keys_pressed=None) -> list[int]:
        """
        - Given input from the simulation make a decision
        :param wall_collision: whether the car collided with the wall
        :param reward_collision: whether the car collided with a reward
        :param inputs: sensor input as a numpy array
        :param keys_pressed: a map of pressed keys
        :return direction: int [0 - num_outputs)
        """
        pass

    @abstractmethod
    def save_model(self, path):
        """
        - Save the brain of the agent to some file (or don't)
        :param path: the path to the model
        :return: None
        """
        pass

    @abstractmethod
    def load_model(self, path):
        """
        - Load the brain of the agent from some file (or don't)
        :param path: the path to the model
        :return: None
        """
        pass



from abc import ABC, abstractmethod

import numpy as np
import pygame
from pygame import image, draw
from pygame.sprite import Sprite
from pygame import transform
from sklearn.preprocessing import normalize

from agent import Agent
from vector2d import Vector2D, Velocity, Angle
from simulation import Simulation


class Vehicle(ABC):
    """
    The car class is responsible for handling the inputs
    and moving through the simulation
    """

    def __init__(self,
                 num_outputs: int,
                 image_path: str = None,
                 driver: Agent = None,
                 scale: int = 1,
                 debug: bool = False,
                 max_speed: int = 20,
                 ignore_max_speed: bool = False,
                 normalize: bool = True):
        # public attributes
        super().__init__()
        self.num_outputs = num_outputs
        self.driver = driver
        self.death_count = 0
        self.current_action = []
        self.sensors: list[Sensor] = []
        # private attributes
        self._normalize = normalize
        self._image_path = image_path
        # cache each rotation upon initialization to increase efficiency
        self._image_angle_cache = []
        self._scale = scale
        self._debug = debug
        # vehicle info
        self.velocity = Velocity(x=0, y=0, angle=0)
        self.odometer = 0
        self.max_speed = max_speed
        self.ignore_max_speed = ignore_max_speed
        self.current_image = None
        self.image = None
        self.init_car_image()

    def _get_vehicle_input(self):
        """
        Overload to add other inputs
        :return: a numpy array of the values from the sensors
        """
        np_array = np.array([sensor.value for sensor in self.sensors])
        norm = np.linalg.norm(np_array)
        return np_array/norm if self._normalize else np_array

    def print_sensor_values(self):
        for i, sensor in enumerate(self.sensors):
            print("Sensor {}:".format(i), sensor.value)

    def init_car_image(self):
        self.image = image.load(self._image_path)
        self.current_image = image
        self.configure_image()
        # todo optimize image processing
        # for i in range(359):
        #     self._image_angle_cache.append(self.rotate_center(i))

    # Unimplemented Feature
    def rotate_center(self, angle):
        """
        Called at initialization to populate the self._image_angle_cache with rotated images
        :param angle: the angle to add to cache
        :return: None
        """
        rotated_image = transform.rotate(self.image, angle)
        new_rect = rotated_image.get_rect(center=self.image.get_rect(topleft=self.rect.topleft).center)
        self._image_angle_cache.append(rotated_image)

    def blit_rotate_center(self, window, top_left):
        """
        Rotate the image at the center and blit to the 'window' surface
        :param window: pygame.surface.Surface
        :param top_left: the top left of the vehicle
        :return: None
        """
        # TODO implement cache using angle
        rotated_image = transform.rotate(self.image, self.velocity.angle.value).convert_alpha()
        new_rect = rotated_image.get_rect(center=self.image.get_rect(topleft=top_left).center)
        self.current_image = rotated_image
        x, y = new_rect.topleft[0] + self.image.get_width() / 2, new_rect.topleft[1] + self.image.get_height() / 2
        window.blit(rotated_image, (x, y))

    def update(self, simulation: Simulation):
        """
        update necessary data and send back to simulation
        :param simulation: the simulation the vehicle exists in
        :return: None
        """
        window = simulation.window
        # account for reoccurring events (such as velocity update)
        self.update_pos()
        self.odometer += self.velocity.speed  # update odometer as the distance moved each step
        self.blit_rotate_center(window, (self.velocity.x, self.velocity.y))

    def update_pos(self):
        self.velocity.transform()

    @staticmethod
    def scale(img, factor):
        nw = img.get_rect().width * factor
        nh = img.get_rect().height * factor
        n_image = transform.scale(img, (nw, nh))
        return n_image

    def init_sensors(self, sensors):
        """
        - Initializes the sensors of the vehicle
        :param sensor_builder:
        :param sensors:
        :return:
        """
        self.sensors = sensors

    def update_sensors(self, window, simulation: Simulation):
        """
        Calculate the values for the sensors of the vehicle
        :param window: the pygame window
        :param simulation: the simulation that the car is functioning in
        :return: None
        """
        # get (x1,y1,x2,y2) tuples for all sensor positions
        s = [sensor.update(window=window, simulation=simulation, car=self) for sensor in self.sensors]
        if self._debug:
            self.display_sensor(car_pos=simulation.get_vehicle_image_position(), window=window)

    # TODO: Display cannot just pull from one collision point
    def display_sensor(self, car_pos, window: pygame.surface.Surface):
        for s in self.sensors:
            if s.coords is not None:
                pygame.draw.line(surface=window,
                                 color=s.line_color,
                                 start_pos=(car_pos),
                                 end_pos=(s.coords[0] if s.collision_point is None else s.collision_point[0],
                                          s.coords[1] if s.collision_point is None else s.collision_point[1]),
                                 width=s.line_width)

    @abstractmethod
    def configure_image(self):
        """
        - perform transformations to vehicle
        - configure 'zero-angle-position' and sizing
        :return:
        """
        pass

    @abstractmethod
    def accelerate(self):
        """
        Make the car go faster, or don't
        :return:
        """
        pass

    @abstractmethod
    def brake(self):
        """
        Make the car go slower, or don't
        :return:
        """
        pass

    @abstractmethod
    def turn(self, left=False, right=False):
        """
        - For custom operation override this method
        rotate the vehicle and update the image (update the velocity 'angle' attribute)
        :param left: whether the car is turning left
        :param right: whether the car is turning right
        :return: None
        """
        pass

    @abstractmethod
    def save_car(self):
        """
        - Should save the model of the driver and any other important information
        :return: None
        """
        pass

    @abstractmethod
    def reset(self, simulation: Simulation):
        """
        - Resets the car properties, so it is ready for another episode
        :return: None
        """
        pass

    @abstractmethod
    def step(self, reward: bool, collision: bool, keys_pressed):
        """
        - Given the reward, collision info and the current input from the sensors, move the car
        :param keys_pressed: pygame keys pressed
        :param reward: whether the car is over a reward
        :param collision: whether the car is over a collision
        :return: None
        """
        pass

    @abstractmethod
    def get_external_inputs(self):
        """
        :return: the number of inputs (not including sensors) that are used
        i.e. velocity
        """
        pass


class SensorBuilder:

    def __init__(self, depth: int, sim: Simulation, default_value=None, color=(0, 0, 0), width=2, pointer=True):
        """
        Convenience class for building sensors
        :param depth: the sensor depth
        :param color: the color of the sensor
        :param width: the width of the sensor
        :param pointer: whether the sensor has an end point
        """
        self.simulation = sim
        self.depth = depth
        self.color = color
        self.width = width
        self.pointer = pointer
        self.default_value = default_value
        self.num_sensors = -1
        self.masks = []  # list of 359 masks representing all angles
        self._sensor_position = np.zeros((360, self.depth, 2))  # list of 359 sensor positions
        self.sensor_depth_array = np.array([self.depth, self.depth])
        self.mask_debug_images = []
        self.offset = 0, 0
        self.generate_sensor_points()

    def get_sensor_pos(self, car, car_angle, sensor_angle):
        """

        :param car: the car object
        :param car_angle: the angle of the vehicle
        :param sensor_angle: the angle of the sensor relative to the car
        :param offset:
        :return: An array of shape ([sensor_depth], 2)
        """
        # Get the sensor position given the current angle and position of the car
        sensor_absolute_angle = car_angle + sensor_angle
        sensor_pos = self._sensor_position[sensor_absolute_angle.round().value]
        return sensor_pos + car.get_vehicle_image_position()

    def get_sensor_offset(self):
        """
        :return: get offset for sensor masks
        """
        car_v = self.simulation.car.velocity
        return (car_v.x - self.depth + self.simulation.car.image.get_width() / 2,
                car_v.y - self.depth + self.simulation.car.image.get_height() / 2)

    def generate_sensors(self, sensor_angles: list[int] = None, sensor_range=None, num_sets=1) -> list:
        """
        - Generate a cached list of sensors
        :param sensor_range: an optional parameter allowing for sensor values to be defined by a range
        :param sensor_angles - list of angles for sensors
        :param range - (start_angle, end_angle, step)
        """
        sensors = []
        bin = []
        if range is not None and sensor_angles is None:
            sensor_angles = range(*sensor_range)

        for i in range(num_sets):
            color = self.color
            if self.color == "random":
                color = self.generate_random_color()
            for angle in sensor_angles:
                self.num_sensors = len(sensor_angles)
                sensors.append(
                    Sensor(
                        sb=self, sensor_depth=self.depth, angle=angle, default_val=self.default_value,
                        offset=np.array(self.offset), line_width=self.width, line_color=color, pointer=self.pointer
                    )
                )
            bin.append(sensors)
            sensors = []
        # Return the list of sensors, or if it's a batch of sensors, return the whole batch
        return bin[0] if num_sets == 1 else bin

    def generate_random_color(self):
        r = np.random.rand() * 255
        g = np.random.rand() * 255
        b = np.random.rand() * 255
        return int(r), int(g), int(b)

    def generate_sensor_points(self):
        """
        - Generate a cached list of sensor masks to be used to receive input
        """
        for angle in range(360):
            a = Angle.create(angle)
            for step in range(self.depth):
                x, y = Vector2D.point_at_angle(step, angle=a, offset=np.array(self.offset))
                x2, y2 = x, -y
                self._sensor_position[angle, step] = np.array([x2, y2])


class Sensor:

    def __init__(self, sb: SensorBuilder, sensor_depth: int,
                 angle: int,
                 default_val: int = None,
                 offset: np.array = np.array([0, 0]),
                 line_width: int = 2,
                 line_color: (int, int, int) = (255, 0, 0),
                 pointer: bool = False
                 ):
        """
        :param sb: SensorBuilder that created the sensor (issue #12)
        :param sensor_depth: The length of the sensor
        :param angle: the angle (relative to the agent)
        :param default_val: the value of the sensor if the sensor doesn't encounter an object.
                            If left None, the default_val param is set to the sensor_depth.
        :param offset: the value to offset the creation of the sensor (assuming the sensor will originate from the
                       top-left corner of the vehicle)
        :param line_width: the width of the sensor line when visible
        :param line_color: the color of the sensor line when visible (default color RED)
        """
        self.sensor_builder = sb
        self.sensor_depth = sensor_depth
        self.angle = angle
        self.default_val = default_val
        if default_val is None:
            self.default_val = self.sensor_depth
        # surface will be initialized later
        self.surf = None
        self.offset = offset
        self.line_width = line_width
        self.line_color = line_color
        self.value = 1  # sensor_depth / sensor_depth
        self.coords = np.zeros(2)
        self.pointer = pointer
        self.collision_point = None

    def update(self, window: pygame.surface.Surface, simulation: Simulation, car):
        """
        Update the surface and the value of the sensor
        :param window: pygame window
        :param simulation: Simulation
        :return: the line coordinates as a tuple
        """

        # update the value of the surface
        surface = pygame.surface.Surface(window.get_size())
        # coords has a shape of ([sensor_depth], 2)
        sensor_array = self.sensor_builder.get_sensor_pos(car, car.velocity.angle, Angle(self.angle))

        # update the value of the sensor
        self.update_value(sensor_array=sensor_array, simulation=simulation, car=car)
        # np.array[x2, y2]
        self.coords = sensor_array[self.sensor_depth - 1]
        # return the line coordinates
        # x2, y2
        return self.coords[0], self.coords[1]

    def update_value(self, sensor_array, simulation, car):
        """
        Check for collisions between the simulation track_border and the sensor surface
        and update the value of the sensor as either the default_val (no bit) or
        the distance between the bit and the vehicle
        Note: Error will be thrown if the track_border is None
        :param car:
        :param sensor_array:
        :param simulation: Simulation
        :return: None
        """
        car_v: Velocity = car.velocity
        border_mask = simulation.border_mask
        angle = car_v.angle + Angle(self.angle)
        # print("car angle: ", car_v.angle)
        # print("mask angle: ", angle)
        for point in sensor_array:
            valid_point = 0 <= point[0] < simulation.track_border_width \
                          and 0 <= point[1] < simulation.track_border_height
            if not valid_point: continue
            bit = border_mask.get_at((point[0], point[1]))
            if bit == 1:
                if self.pointer:
                    x, y = point[0], point[1]
                    draw.circle(simulation.window, (255, 255, 255), (x, y), 5)
                self.value = car_v.distance_between(
                    other=Vector2D(x=point[0], y=point[1]),
                    offset=(car.image.get_width() / 2,
                            car.image.get_height() / 2)
                )
                self.collision_point = point
                break
            else:
                self.value = self.default_val
                self.collision_point = None



from setuptools import setup, find_packages

classifiers = []

with (open("README.md")) as f: 
  long_description = f.read()
 

setup(
  name="pysimulate",
  url="https://github.com/lukeharwood11/ai-racecar-simulator",
  author="Luke Harwood",
  author_email="mwstudiodev@gmail.com",
  version="1.0.3",
  description="AI Track Simulator is an easy-to-use Simulator engine built in pygame",
  long_description=long_description,
  packages=["simulator"],
  keywords='simulator ai-gym racecar',
  package_dir={'': 'org.mwdev'},
  include_package_data=True,
  license="MIT",
  classifiers=classifiers,
  command_options={},
  python_requires=">=3.6",
  install_requires=[
    'numpy', 'tensorflow', 'pygame', 'keras', 'sklearn', 'scikit-learn'
  ]
)



import os
import time

import numpy as np
from gui.components import TimedLabel, TimedLabelQueue
from tensorflow.keras.optimizers import Adam
from agent import Agent
from tensorflow.keras import Sequential
from tensorflow.keras.layers import Dense, InputLayer
from pygame import (K_UP, K_DOWN, K_LEFT, K_RIGHT, transform)


class Experience:

    def __init__(self, current_state, current_action, resulting_reward, resulting_state):
        """
        Holds the memory for replay
        :param current_state: the current state of the model
        :param current_action: the action that was chosen
        :param res_reward: the resulting reward
        :param res_state: the resulting state
        """
        self.current_state = current_state
        self.current_action = current_action
        self.res_reward = resulting_reward
        self.next_state = resulting_state


class ReplayMemory:
    """
    - Serves as a queue that has a max_size
    - Max size is by default None and should be added by the user
    """

    def __init__(self, max_size=None):
        self._max_size = max_size
        self._arr = []
        self._size = 0

    def add_experience(self, ex: Experience):
        if self._size == self._max_size and self._max_size is not None:
            self._arr.pop(0)
            self._size -= 1
        self._arr.append(ex)
        self._size += 1

    def get_random_experiences(self, num_samples):
        np.random.shuffle(self._arr)
        return self._arr[0:num_samples if num_samples < self._size else self._size]


class QLearningParams:
    def __init__(self, wall_collision_value, reward_collision_value, other_value, timeout_value):
        self.wall = wall_collision_value
        self.reward = reward_collision_value
        self.other = other_value
        self.timeout = timeout_value

    def speed_reward(self, current_reward_value, speed):
        """
        reward the model for going faster, and punish for going too slow
        :return:
        """
        return current_reward_value + speed/4


class QLearningAgent(Agent):

    def __init__(self, simulation, alpha, alpha_decay, y, epsilon, num_sensors, num_actions, batch_size, replay_mem_max,
                 save_after=None,
                 load_latest_model=False, training_model=True, model_path=None, train_each_step=False, debug=False, other_inputs=0, timeout=None):
        # initialize Agent parent class
        # add one to num_inputs for current speed
        super(QLearningAgent, self).__init__(num_inputs=num_sensors+other_inputs, num_outputs=num_actions)
        self.simulation = simulation
        # Q learning hyperparameters
        self.alpha = alpha
        self.alpha_decay = alpha_decay
        self.y = y
        self.epsilon = epsilon
        self.batch_size = batch_size
        # Q learning replay memory
        self.replay_memory = ReplayMemory(max_size=replay_mem_max)
        # private state
        self._last_reward_time = time.time()
        self._timeout = timeout
        self._current_state = None
        self._current_action = None
        self._rewarded_currently = False
        self._collision_count = 0
        # load/save/training properties
        self._save_after = save_after
        self._load_latest_model = load_latest_model
        self._training_model = training_model  # boolean
        self._model_path = model_path
        self._train_each_step = train_each_step
        # debug private attributes
        self._debug = debug
        # build Sequential tensorflow model
        self._model = Sequential()
        self._model.add(InputLayer(input_shape=self.num_inputs))
        self._model.add(Dense(64))
        self._model.add(Dense(32, activation='tanh'))
        self._model.add(Dense(self.num_outputs))
        self._model.compile(loss='mse', optimizer=Adam(learning_rate=self.alpha))
        self._model.summary()

        if self._model_path is not None:
            self.load_model(self._model_path)
        elif self._load_latest_model:
            self.init_default_model_weights()
        # Q learning rewards
        self._qlearn_params = QLearningParams(wall_collision_value=-20, reward_collision_value=20, other_value=-2, timeout_value=-10)

    def update(self, inputs, reward_collision=False, wall_collision=False, keys_pressed=None) -> list[int]:
        """
        - Given input from the simulation make a decision
        :param wall_collision: whether the car collided with the wall
        :param reward_collision: whether the car collided with a reward
        :param inputs: sensor input as a numpy array
        :param keys_pressed: a map of pressed keys (ignore, n/a)
        :return direction: int [0 - num_outputs)
        """
        reward = self._get_reward(reward_collision, wall_collision)
        # Change internal states
        self._handle_collision(wall_collision)
        reward = self._handle_reward(reward, reward_collision)
        # If the user presses the down key- restart the simulation
        if keys_pressed[K_DOWN] and self._training_model:
            reward = self._handle_restart(reward)
        # If the user presses the up key- simply train the model
        if keys_pressed[K_UP] and self._training_model:
            self._train_model()
        reward = self._qlearn_params.speed_reward(reward, self.simulation.car.velocity.speed)
        self._handle_experience(reward, inputs)
        self._handle_training()
        actions = self._model.predict(inputs.reshape(1, self.num_inputs))
        action = np.argmax(actions)
        if np.random.rand() > self.epsilon and self._training_model:
            action = np.random.choice(np.arange(self.num_outputs))
        self._current_action = action
        print("Current Action:", action, "Current Reward:", reward, "Choices:", actions)
        return [action]

    def _get_reward(self, reward_collision, wall_collision):
        if wall_collision:
            return self._qlearn_params.wall
        elif reward_collision:
            return self._qlearn_params.reward
        else:
            return self._qlearn_params.other

    def _handle_experience(self, reward, inputs):
        if self._current_state is not None:
            self.replay_memory.add_experience(Experience(
                current_state=self._current_state,
                current_action=self._current_action,
                resulting_reward=reward,
                resulting_state=inputs
            ))
        self._current_state = inputs

    def _handle_training(self):
        if self._training_model:
            if self._train_each_step:
                self._train_model()

    def _handle_collision(self, wall_collision):
        if wall_collision:
            if self._collision_count % self._save_after == 0:
                self._save_model_increment()
            if self._training_model:
                self._train_model()
            self._collision_count += 1

    def _handle_reward(self, reward, reward_collision):
        if reward_collision:
            if self._rewarded_currently:
                # if the car is sitting on a reward, punish it with "other" value
                reward -= self._qlearn_params.reward - self._qlearn_params.other
            else:
                self._last_reward_time = time.time()
                if self._debug:
                    self.simulation.label_manager.display_label(TimedLabel(
                        position=(550, 600), timeout=2.0, queue=self.simulation.label_manager, text="Reward Found!", size=30,
                        font=None, color=(255, 255, 255), refresh_count=None, background=(0, 0, 0), anti_alias=True
                    ), force=True)
            self._rewarded_currently = True
        else:
            self._rewarded_currently = False
            if self._timeout is not None and time.time() - self._last_reward_time >= self._timeout:
                reward = self._handle_restart(reward)
        return reward

    def _handle_restart(self, reward):
        self._last_reward_time = time.time()
        self._request_restart()
        self._collision_count += 1
        reward += self._qlearn_params.timeout
        self._train_model()
        self.simulation.label_manager.display_label(TimedLabel(
            position=(400, 250), timeout=2.0, queue=self.simulation.label_manager, text="Timeout! Restart!", size=80,
            font=None, color=(255, 255, 255), refresh_count=None, background=(0, 0, 0), anti_alias=True
        ), force=False)
        return reward

    def _request_restart(self):
        self.simulation.reset()

    def _train_model(self):
        """
        - Get [self.batch_size] number of experiences and train on those experiences
        """
        batch = self.replay_memory.get_random_experiences(self.batch_size)
        X_train = np.zeros((self.batch_size, self.num_inputs))
        y_train = np.zeros((self.batch_size, self.num_outputs))
        for i, experience in enumerate(batch):
            current_state = np.array(experience.current_state).reshape(1, self.num_inputs)  # TODO reshape
            # predict the q_values
            q_value_prediction = self._model.predict(current_state)
            # set the target to be what the experience actually was
            q_target = experience.res_reward
            resulting_state = np.array(experience.next_state)  # TODO reshape
            q_target = q_target + self.y * np.max(self._model.predict(resulting_state.reshape(1, self.num_inputs))[0])
            # adjust the weights (no other q_vals are impacted)
            q_value_prediction[0][experience.current_action] = q_target
            X_train[i] = current_state
            y_train[i] = q_value_prediction[0]

        self._model.fit(X_train, y_train, verbose=2, epochs=3, batch_size=16)

    def _save_model_increment(self):
        """
        Save the current model to a unique location representing the current iteration
        :return: None
        """
        self._model.save_weights("assets/models/model_" + str(self._collision_count) + ".h5")

    def save_model(self, path):
        """
        - Save the brain of the agent to some file (or don't)
        :param path: the path to the model
        :return: None
        """
        self._model.save_weights(os.path.join(path, 'latest.h5'))

    def init_default_model_weights(self):
        self._model.load_weights(os.path.join("assets", "models", "latest.h5"))

    def load_model(self, path):
        """
        - Load the brain of the agent from some file (or don't)
        :param path: the path to the model
        :return: None
        """
        self._model.load_weights(os.path.join("assets", "models", path))



from math import sin, cos, tan, sqrt, radians, degrees
import numpy as np


class Angle:

    def __init__(self, starting_val=0):
        self.value = starting_val

    def update(self, dA):
        self.value = self._calc_difference(dA)

    def _calc_difference(self, dA):
        """
        - other: Angle
        """
        value = self.value
        value += dA
        return value % 360

    def greater_than_180(self):
        return self.value > 180

    def ref(self):
        if self.value >= 180:
            return Angle((360 - self.value) % 360)
        return self

    def __add__(self, other):
        diff = self.value + other.value
        return Angle(diff % 360)

    def __sub__(self, other):
        diff = self.value - other.value
        return Angle(self._calc_difference(diff))

    def __eq__(self, other):
        return other.value == self.value

    def __mul__(self, other):
        return Angle(other.value * self.value)

    def __lt__(self, other):
        return self.value < other

    def __gt__(self, other):
        return self.value > other

    def __le__(self, other):
        return self.value <= other

    def __ge__(self, other):
        return self.value >= other

    def __str__(self):
        return "Value: {0}".format(self.value)

    def to_radians(self):
        return radians(self.value)

    def sin(self):
        return sin(radians(self.value))

    def cos(self):
        return cos(radians(self.value))

    def tan(self):
        return tan(radians(self.value))

    def round(self):
        v = round(self.value)
        return Angle(v if v != 360 else 0)

    @staticmethod
    def create(angle):
        return Angle(angle % 360)


class Vector2D:
    """
    - Convenience class to hold 2d data
    """

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def get_magnitude(self):
        return sqrt(self.x ** 2 + self.y ** 2)

    @staticmethod
    def calc_magnitude(x, y):
        return sqrt(x ** 2 + y ** 2)

    def distance_between(self, other, offset=np.array([0, 0])) -> float or None:
        """
        :param other: Vector
        :param offset: offset of current vector
        :return: the distance between the two points or None if other is None
        """
        if other is None:
            return None
        else:
            dy = other.y - (self.y + offset[1])
            dx = other.x - (self.x + offset[0])
            return Vector2D.calc_magnitude(dx, dy)

    @staticmethod
    def point_at_angle(distance, angle: Angle, offset=np.array([0, 0])):
        """
        Projects a point [distance] length away at an [angle] relative to the current direction of the vector
        :param distance: distance of the point
        :param angle: angle relative to the current angle of the car
        :param offset: the offset of the position to account for rect positioning in pygame
        :return: (x:int, y:int) -> tuple of the coordinates of the projected position
        """
        return np.array([angle.cos() * distance, angle.sin() * distance]) + offset


class Velocity(Vector2D):

    def __init__(self, speed=0, x=0, y=0, angle=0):
        """
        :param x: the current x position
        :param y: the current y position
        :param angle: the current angle the object is moving in
        :param speed: the current speed
        """
        super(Velocity, self).__init__(x, y)
        self.angle = Angle(starting_val=angle)
        self.speed = speed

    def turn(self, da):
        """
        :param da: the change in the angle
        :return: None
        """
        self.angle.update(da)

    def transform(self):
        """
        :param distance:
        :return:
        """
        dx = cos(radians(self.angle.value)) * self.speed
        dy = -sin(radians(self.angle.value)) * self.speed
        self.x += dx
        self.y += dy

    def reset_velocity(self, x=0, y=0, angle=0, speed=0):
        self.x = x
        self.y = y
        self.angle = Angle.create(angle)
        self.speed = speed



from abc import ABC

import numpy as np

from agent import Agent
from simulation import Simulation
from vehicle import Vehicle, SensorBuilder
from qlearn import QLearningAgent
from pygame import (K_UP, K_DOWN, K_LEFT, K_RIGHT, transform)
from gui.components import TimedLabel, Label
import os


class DefaultSimulation(Simulation, ABC):

    def __init__(self,
                 debug=True,
                 fps=None,
                 num_episodes=None,
                 caption: str = None,
                 car: Vehicle = None,
                 track_offset=(0, 0),
                 screen_size=(1400, 800),
                 track_size=(1400, 800)):
        super(DefaultSimulation, self).__init__(debug=debug,
                                                fps=fps,
                                                num_episodes=num_episodes,
                                                caption=caption,
                                                car=car,
                                                track_offset=track_offset,
                                                screen_size=screen_size,
                                                track_size=track_size)

    def init_car_start_pos(self):
        """
        sets the start position of the car
        :return:
        """
        self.start_pos = (875, 100)

    def update_and_display_arrow_display(self):
        self.arrow_display.render(
            self.window, (520, 500), current_actions=self.car.current_action
        )

    def init_track(self) -> (str, str, str):
        """
        Should set the images of the track (paths to the images):
        called in the constructor of the simulation class
        - track border
        - track bg
        - track rewards
        :return: the path to the tracks in the order 'border, background (design), rewards'
        """
        return \
            os.path.join("assets", "track-border.png"), \
            os.path.join("assets", "track.png"), \
            os.path.join("assets", "track-rewards.png")


class Car(Vehicle, ABC):

    def __init__(self, driver, debug=False, acceleration_multiplier=.5, normalize=True):
        super(Car, self).__init__(
            num_outputs=5,
            image_path=os.path.join("assets", "grey-car.png"),
            driver=driver,
            scale=1,
            debug=debug,
            max_speed=20,
            normalize=normalize
        )
        self.odometer_label = None
        self.speed_label = None
        self.acceleration_multiplier = acceleration_multiplier
        self.model_path = os.path.join("assets", "models")

    @staticmethod
    def get_num_outputs():
        """
        0 = left
        1 = accelerate
        2 = right
        3 = break
        4 = coast
        :return: number of inputs
        """
        return 5

    def get_vehicle_image_position(self):
        """
        :return: The absolute position of the image of the vehicle (in relation to the window)
        """
        return np.array((self.velocity.x + (self.image.get_width() / 2) + 12,
                  self.velocity.y + (self.image.get_height() / 2) + 12))

    def configure_image(self):
        self.image = transform.rotate(self.image, -90)
        self.image = transform.smoothscale(self.image, (34, 17))

    def save_car(self):
        """
        - Should save the model of the driver and any other important information
        :return: None
        """
        if self.driver is not None:
            if not os.path.exists(self.model_path):
                os.mkdir(self.model_path)
            self.driver.save_model(self.model_path)

    def reset(self, simulation):
        """
        - Resets the car properties, so it is ready for another episode
        :return: None
        """
        self.velocity.reset_velocity(
            x=simulation.start_pos[0],
            y=simulation.start_pos[1],
            angle=180,
            speed=0
        )
        self.odometer = 0

    def accelerate(self):
        """
        Accelerate the car
        :return: None
        """
        if self.ignore_max_speed or self.velocity.speed < self.max_speed:
            self.velocity.speed += self.acceleration_multiplier

    def turn(self, left=False, right=False):
        if left:
            self.velocity.turn(self.velocity.speed * .6)
        if right:
            self.velocity.turn(self.velocity.speed * -.6)

    def brake(self):
        """
        Slow down the car or stop if the speed is less than a threshold value
        :return: None
        """
        if self.velocity.speed > 1:
            self.velocity.speed -= self.acceleration_multiplier
        else:
            self.velocity.speed = 0

    def step(self, reward: bool, collision: bool, keys_pressed):
        """
        - Given the reward, collision info and the current input from the sensors, move the car
        - In this implementation of step() the inputs consist of the sensor depths with the last two values of inputs
        being collision (input[-2]) and reward (input[-1]) with 1 being a collision/reward True and 0 being False
        :param keys_pressed: pygame keys pressed
        :param reward: whether the car is over a reward
        :param collision: whether the car is over a collision
        :return: None
        """
        i = self._get_vehicle_input()
        direction = self.driver.update(inputs=i, wall_collision=collision, reward_collision=reward,
                                       keys_pressed=keys_pressed)
        self.current_action = direction
        accel = False
        if direction.count(0) > 0:
            self.turn(left=True)
        if direction.count(1) > 0:
            self.accelerate()
            accel = True
        if direction.count(2) > 0:
            self.turn(right=True)
        if direction.count(3) > 0:
            self.brake()
        if not accel:
            self.deccelerate()

    def deccelerate(self):
        self.velocity.speed = .98 * self.velocity.speed

    def get_external_inputs(self):
        """
        :return: 1 for speed
        """
        return 1

    def _get_vehicle_input(self):
        """
        :return: a numpy array of the values from the sensors
        """
        np_array = np.array([sensor.value for sensor in self.sensors] + [self.velocity.speed])
        norm = np.linalg.norm(np_array)
        return np_array/norm if self._normalize else np_array


class GameControlDriver(Agent, ABC):

    def __init__(self, num_inputs, num_outputs):
        """
        Default controls for human driver
        :param num_inputs: the number of inputs (sensors/collision/reward)
        :param num_outputs: the number of outputs (driver-actions/left/right/etc.)
        """
        super().__init__(num_inputs, num_outputs)

    def update(self, inputs, reward_collision=False, wall_collision=False, keys_pressed=None) -> list[int]:
        """
        - Encode the inputs to integers 0 - 3
        :param wall_collision: n/a
        :param reward_collision: n/a
        :param inputs: the input from the car sensors (n/a)
        :param keys_pressed: the keys pressed from the user
        :return: a list of output encodings (0 - 3) representing requested movement
        """
        print(
            "collision", wall_collision,
            "reward_collision", reward_collision,
            "inputs", inputs
        )
        ret = []
        if keys_pressed[K_LEFT]:
            ret.append(0)
        if keys_pressed[K_UP]:
            ret.append(1)
        if keys_pressed[K_RIGHT]:
            ret.append(2)
        if keys_pressed[K_DOWN]:
            ret.append(3)
        return ret

    def save_model(self, path):
        """
        do nothing
        :param path: n/a
        :return: n/a
        """
        pass

    def load_model(self, path):
        """
        do nothing
        :param path: n/a
        :return: n/a
        """
        pass


def main():

    car = Car(
        driver=None,
        debug=False,
        acceleration_multiplier=.5,
        normalize=True
    )

    simulation = DefaultSimulation(
        debug=False,
        fps=35,  # None means simulation fps is not tracked (Suggested for training)
        num_episodes=None,
        caption="Default Simulation",
        car=car,
        track_offset=(0, 0),
        screen_size=(1400, 800),
        track_size=(1400, 800)
    )
    # Create Sensors
    sb = SensorBuilder(
        sim=simulation,
        depth=500,
        default_value=None,
        color=(255, 0, 0),
        width=2,
        pointer=True
    )
    sensors = sb.generate_sensors(sensor_range=(-90, 90, 5))

    driver_map = {
        'user': GameControlDriver(
            num_inputs=len(sensors),
            num_outputs=car.num_outputs
        ),
        'qlearn': QLearningAgent(
            simulation=simulation,
            alpha=0.01,
            alpha_decay=0.01,
            y=0.90,
            epsilon=.98,
            num_sensors=len(sensors),
            num_actions=car.num_outputs,
            batch_size=64,
            replay_mem_max=400,
            save_after=100,
            load_latest_model=True,
            training_model=False,
            model_path=None,
            train_each_step=False,
            debug=False,
            other_inputs=car.get_external_inputs(),
            timeout=10
        )
    }

    # Change this line to select different drivers
    driver = driver_map['qlearn']

    # sensors = sb.generate_sensors([0])
    # Attach sensors to car
    car.init_sensors(sensors=sensors)
    # Throw driver in the vehicle
    car.driver = driver
    simulation.simulate()


if __name__ == "__main__":
    main()
